<div class="refresh-trigger accordion-heading-button btn-group" data-bind="visible: $root.refreshVisible">
    <a href="#" data-bind="click: function() { $root.requestData(); }" title="{{ _('Refresh connection options')|edq }}">
        <i class="fa fa-refresh"></i>
    </a>
</div>
