#settings_plugin_softwareupdate_updatelist {
  display: block;
  width: auto;
  max-height: 500px;
  overflow-x: hidden;
  overflow-y: scroll;
  padding-right: 2px;
  margin-bottom: 1em;
}

td.settings_plugin_softwareupdate_column_update {
  width: 16px;
}

td.settings_plugin_softwareupdate_column_information {
  .line {
    display: block;
  }
}

#settings_plugin_softwareupdate_workingdialog_output {
  font-size: 0.8em;

  .message {
    font-weight: bold;
  }

  .separator {
    font-weight: bold;
    color: #666666;
  }

  .stdout {
    color: #333333;
  }

  .stderr {
    color: #990000;
  }

  .call {
    color: #000099;
  }

  .message_error {
    .message;
    .stderr;
  }
}

.softwareupdate_notification {
  ul {
    margin: 10px 0 10px 25px;

    .name {
      text-overflow: ellipsis;
      white-space: nowrap;
      overflow: hidden;
      display: block;
    }
  }
}

