<div class="control-group" title="{{ _('Video codec used for encoding') }}">
    <label class="control-label" for="settings-webcam_ffmpegVideoCodec">{{ _('Video codec') }}</label>
    <div class="controls">
        <select data-bind="options: webcam_available_videocodecs, value: webcam_ffmpegVideoCodec"></select>
    </div>
</div>
