<h3>{{ _('Printer Profiles') }}</h3>

{% include "snippets/settings/printerprofiles/profiles.jinja2" %}

<button class="btn pull-right" data-bind="click: function() { $root.printerProfiles.showEditProfileDialog(); }"><i class="fa fa-plus"></i> {{ _('Add Profile...') }}</button>

{% include "snippets/settings/printerprofiles/profileEditor.jinja2" %}
