<div class="control-group" title="{{ _('URL to embed into the UI for live viewing of the webcam stream')|edq }}">
    <label class="control-label" for="settings-webcamStreamUrl">{{ _('Stream URL') }}</label>
    <div class="controls">
        <div class="input-append">
            <input type="text" class="input-block-level" data-bind="value: webcam_streamUrl, valueUpdate: 'afterkeydown'" id="settings-webcamStreamUrl">
            <button class="btn" type="button" data-bind="click: testWebcamStreamUrl, enable: webcam_streamUrl() && !testWebcamStreamUrlBusy(), css: {disabled: !webcam_streamUrl() || testWebcamStreamUrlBusy()}"><i class="fa fa-spinner fa-spin" data-bind="visible: testWebcamStreamUrlBusy"></i> {{ _('Test') }}</button>
        </div>
        <span class="help-inline">{% trans %}Needs to be reachable from the browser displaying the OctoPrint UI, used to embed the webcam stream into the page.{% endtrans %}</span>
    </div>
</div>
