<h3>{{ _('Webcam') }}</h3>

<form class="form-horizontal">
    {% include "snippets/settings/webcam/webcamEnabled.jinja2" %}
    {% include "snippets/settings/webcam/webcamStreamUrl.jinja2" %}
    {% include "snippets/settings/webcam/webcamStreamRatio.jinja2" %}
    {% include "snippets/settings/webcam/webcamOrientation.jinja2" %}

    <div>
        <div><small><a href="#" class="muted" data-bind="toggleContent: { class: 'fa-caret-right fa-caret-down', parent: '.form-horizontal', container: '.hide' }"><i class="fa fa-caret-right"></i> {{ _('Advanced options') }}</a></small></div>
        <div class="hide">
            {% include "snippets/settings/webcam/webcamStreamTimeout.jinja2" %}
        </div>
    </div>
</form>

<h3>{{ _('Timelapse Recordings') }}</h3>

<form class="form-horizontal">
    {% include "snippets/settings/webcam/timelapseEnabled.jinja2" %}
    {% include "snippets/settings/webcam/webcamSnapshotUrl.jinja2" %}
    {% include "snippets/settings/webcam/ffmpegPath.jinja2" %}
    {% include "snippets/settings/webcam/watermark.jinja2" %}
    <div>
        <div><small><a href="#" class="muted" data-bind="toggleContent: { class: 'fa-caret-right fa-caret-down', parent: '.form-horizontal', container: '.hide' }"><i class="fa fa-caret-right"></i> {{ _('Advanced options') }}</a></small></div>
        <div class="hide">
            {% include "snippets/settings/webcam/ffmpegBitrate.jinja2" %}
            {% include "snippets/settings/webcam/ffmpegThreads.jinja2" %}
            {% include "snippets/settings/webcam/webcamSnapshotTimeout.jinja2" %}
            {% include "snippets/settings/webcam/webcamSnapshotSslValidation.jinja2" %}
            {% include "snippets/settings/webcam/ffmpegVideoCodec.jinja2" %}
        </div>
    </div>
</form>
