# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

ADMIN_GROUP = "admins"
USER_GROUP = "users"
GUEST_GROUP = "guests"
READONLY_GROUP = "readonly"
