<h3>{{ _('Error Tracking') }}</h3>

{% trans %}
<p>
    Error tracking will cause any logged exceptions in the server and the browser interface to be sent
    to OctoPrint's <a href="https://sentry.io/" target="_blank" rel="noopener noreferer">Sentry instance</a>.
</p>
<p>
    By enabling it you help to gather detailed information on the cause of bugs or other issues. This is especially valuable
    on release candidates, which is why this plugin will also prompt you to enable error tracking if it detects that
    you are subscribed to an RC release channel.
</p>
<p>
    Please enable this if you are running a release candidate, or are prompted to do so while debugging an issue that you reported.
</p>
{% endtrans %}

<form class="form-horizontal">
    <div class="control-group">
        <div class="controls">
            <label class="checkbox">
                <input type="checkbox" data-bind="checked: settings.plugins.errortracking.enabled"> {{ _('Enable error tracking') }}
            </label>
            <span class="help-block">{{ _('Please reload the page after enabling/disabling & saving') }}</span>
        </div>
    </div>
    <div class="control-group">
        <label class="control-label" title="{{ _('Random unique instance ID')|edq }}">{{ _('Instance ID') }}</label>
        <div class="controls">
            <span style="padding-top: 5px; margin-bottom: 5px; display: inline-block" data-bind="text: settings.plugins.errortracking.unique_id() || 'N/A'"></span>
        </div>
    </div>
</form>

{% trans %}
<p>
    The Error Tracking plugin is using a Sentry instance kindly provided by
    <a href="https://sentry.io/" target="_blank" rel="noopener noreferer">sentry.io</a>. For information on their service
    please refer to their <a href="https://sentry.io/security/" target="_blank" rel="noopener noreferer">Security & Compliance documentation</a>
    and their <a href="https://sentry.io/privacy/" target="_blank" rel="noopener noreferer">Privacy Policy</a>.
</p>
{% endtrans %}
