<p>{% trans %}
    If the connectivity check is enabled, OctoPrint will regularly check if it's connected to the internet.
    This is <strong>useful to prevent resource intensive operations</strong> (such as checking for updates) if it's already
    clear that they won't succeed anyhow.
{% endtrans %}</p>
