#settings_plugin_appkeys_userkeys_table,
#settings_plugin_appkeys_allkeys_table {
  .settings_plugin_appkeys_app {
  }
  .settings_plugin_appkeys_actions {
    width: 75px;
    text-align: center;

    a {
      color: black;
    }
  }
  .settings_plugin_appkeys_user {
    width: 100px;
  }
  .settings_plugin_appkeys_checkbox {
    text-align: center;
    width: 10px;

    input[type="checkbox"] {
      margin-top: 0;
    }
  }
}
