#sidebar_plugin_printer_safety_check_wrapper {
  border: 1px solid #b94a48;

  .accordion-heading {
    background-color: #b94a48;
    font-weight: bold;
    color: #fff;
    a, a:hover, a:visited, a:active {
      color: #fff;
    }
    .icon-black {
      color: #fff;
    }
  }
}
