# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

__license__ = 'GNU Affero General Public License http://www.gnu.org/licenses/agpl.html'

class FileDestinations(object):

	SDCARD = "sdcard"
	LOCAL = "local"
