#settings_plugin_pluginmanager {
}

#settings_plugin_pluginmanager_pluginlist {
  display: block;
  width: auto;
  max-height: 400px;
  overflow-x: hidden;
  overflow-y: scroll;
  padding-right: 2px;
  margin-bottom: 1em;

  table {

    th, td {
      &.settings_plugin_plugin_manager_plugins_name {
        text-overflow: ellipsis;
        text-align: left;
      }

      &.settings_plugin_plugin_manager_plugins_actions {
        text-align: center;
        width: 80px;

        a {
          text-decoration: none;
          color: #000;

          &.disabled {
            color: #ccc;
            cursor: default;
          }
        }
      }
    }
  }
}

#settings_plugin_pluginmanager_orphanlist {
  display: block;
  width: auto;
  max-height: 400px;
  overflow-x: hidden;
  overflow-y: scroll;
  padding-right: 2px;
  margin-bottom: 1em;

  table {

    th, td {
      &.settings_plugin_plugin_manager_orphans_id {
        text-overflow: ellipsis;
        text-align: left;
      }

      &.settings_plugin_plugin_manager_orphans_actions {
        text-align: center;
        width: 80px;

        a {
          text-decoration: none;
          color: #000;

          &.disabled {
            color: #ccc;
            cursor: default;
          }
        }
      }
    }
  }
}

#settings_plugin_pluginmanager_repositorydialog {
  h4 {
    position: relative;

    a.dropdown-toggle {
      color: inherit;
      text-decoration: none;
      font-size: 14px;
    }

    ul.dropdown-menu {
      font-size: 14px;
    }
  }

  .form-search {
    text-align: center;
    margin-bottom: 5px !important;
  }

  .form-inline {
    padding: 5px;
    padding-right: 10px;
    margin-bottom: 0;

    .help-block {
      margin-bottom: 0;
      font-size: 85%;
    }
  }

  #settings_plugin_pluginmanager_repositorydialog_unavailable,
  #settings_plugin_pluginmanager_repositorydialog_empty {
    overflow: hidden;
    width: 100%;
    height: 300px;
    background-image: url("../img/repo_unavailable.png");
    text-align: center;
    display: table;

    div {
      display: table-cell;
      vertical-align: middle;
    }
  }

  #settings_plugin_pluginmanager_repositorydialog_list {
    width: auto;
    height: 300px;
    overflow-x: hidden;
    overflow-y: scroll;
    padding-right: 2px;

    .entry {
      border-bottom: 1px solid #ddd;
      padding: 5px;
    }
  }
}

#settings_plugin_pluginmanager_workingdialog_output {
  .message {
    font-weight: bold;
  }

  .error {
    font-weight: bold;
    color: #990000;
  }

  .stdout {
    color: #333333;
  }

  .stderr {
    color: #990000;
  }

  .call {
    color: #000099;
  }
}
