<div class="control-group" title="{{ _('Command to restart the OctoPrint server')|edq }}">
    <label class="control-label" for="settings-serverRestartCommand">{{ _('Restart OctoPrint') }}</label>
    <div class="controls">
        <input type="text" class="input-block-level" data-bind="value: server_commands_serverRestartCommand" id="settings-serverRestartCommand">
    </div>
</div>
